from typing import List, Dict, Union
import re, subprocess

Arg = Union[str, int, float]
ListArg = List[Arg]
DictArg = Dict[str, Union[Arg, List[Arg]]]

# forCmd: 将会为含有空格的参数包裹上引号
def toWin(p: str, forCmd=False) -> str:
  winP = re.sub("/", "\\\\", p.strip())
  if forCmd and re.match(r'\S+\s', winP):
    return '"%s"' % winP
  return winP

def callJs(p: str, args: Union[ListArg, DictArg], nodeExePath="node"):
  cmd = [nodeExePath, p]
  if isinstance(args, list):
    cmd.extend(args)
  else:
    for k in args.keys():
      if isinstance(args[k], list):
        for val in args[k]:
          cmd.append("%s" % k)
          cmd.append("%s" % str(val))
      else:
        cmd.append("%s" % k)
        cmd.append("%s" % str(args[k]))
  subprocess.call(" ".join([toWin(s, True) for s in cmd]))

def callPython(p: str, args: Union[ListArg, DictArg], pythonExePath="python"):
  cmd = [pythonExePath, p]
  if isinstance(args, list):
    cmd.extend(args)
  else:
    for k in args.keys():
      if isinstance(args[k], list):
        for val in args[k]:
          cmd.append("%s" % k)
          cmd.append("%s" % str(val))
      else:
        cmd.append("%s" % k)
        cmd.append("%s" % str(args[k]))
  subprocess.call(" ".join([toWin(s, True) for s in cmd]))



from typing import Union, List, Tuple
from smtplib import SMTP_SSL, SMTP_SSL_PORT
from tangUtils.main import File

from email import encoders
from email.header import Header
from email.utils import formataddr, formatdate, COMMASPACE

from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart

# str: ip address
# Tuple[str, str]: (username, ip address)
Addr = Union[str, Tuple[str, str]]

DefaultPlainText = "您的客户端无法显示此类型的消息, 请换用其它客户端以查看该超文本内容(text/html)"

def _encode(s: str) -> str:
  return Header(s, "utf-8").encode()


def _decodeAddr(_addr: Addr) -> Tuple[str, str]:
  """
    _addr:
      str: 表示昵称和地址均为该值
      Tuple[str, str]: 等价于(username, ip address)
    return (username, ip address)
  """
  _type = type(_addr)
  if _type == str:
    name = _addr
    addr = _addr
  elif _type == tuple:
    name, addr = _addr
  else:
    raise Exception("arguments type error, accept str or Tuple[str, str], but got", _addr)
  return (name, addr)


def _formatAddr(_addr: Addr):
  """
    返回格式化后的收/发件人
    类似于  username <ip-address@mail.com>
  """
  name, addr = _decodeAddr(_addr)
  encodedName = _encode(name)
  return formataddr((encodedName, addr))


class Mail(object):
  """
    所有类方法都没有加 try/except, 如有需要则自己添加
  """
  # 发件人昵称
  fromName: str
  # 发件人邮箱地址
  fromAddr: str
  # 发件人邮箱授权码
  password: str
  # SMTP服务器地址, 默认为 qq 邮箱: "smtp.qq.com"
  smtpServer: str
  # SMTP服务器端口, 默认为 SMTP_SSL_PORT
  smtpPort: int
  # SMTP_SSL
  server: SMTP_SSL
  # 登录标记
  loginFlag = False

  def __init__(self, fromAddr: Addr, password: str, smtpServer="smtp.qq.com", smtpPort=SMTP_SSL_PORT, debugLevel=0):
    super().__init__()
    self.fromName, self.fromAddr = _decodeAddr(fromAddr)
    self.password = password
    self.smtpServer = smtpServer
    self.smtpPort = smtpPort
    self.setServer(smtpServer, smtpPort, debugLevel)

  def setServer(self, smtpServer="smtp.qq.com", smtpPort=SMTP_SSL_PORT, debugLevel=0):
    print("正在连接邮箱服务器...")
    self.server = SMTP_SSL(self.smtpServer, self.smtpPort)
    self.loginFlag = False
    self.server.set_debuglevel(debugLevel)
    return self

  def login(self):
    print("正在登录邮箱...")
    self.server.login(self.fromAddr, self.password)
    self.loginFlag = True
    return self

  def quit(self):
    self.server.quit()
    self.loginFlag = False
    return self

  def sendTo(self, toAddr: List[Addr], subject="", fallbackText=DefaultPlainText, html="", attachment: List[File] = []):
    print("正在发送消息...")
    if not self.loginFlag:
      self.login()
    # 如果登录之后, 登录标记依旧为否, 则取消消息发送
    if not self.loginFlag:
      raise Exception("登录失败")
    msg = MIMEMultipart("alternative")
    msg["From"] = _formatAddr((self.fromName, self.fromAddr))
    msg["To"] = COMMASPACE.join([_formatAddr(item) for item in toAddr])
    msg["Date"] = formatdate(localtime=True)
    msg["Subject"] = _encode(subject)

    msg.attach(MIMEText(fallbackText, "plain", "utf-8"))
    msg.attach(MIMEText(html, "html", "utf-8"))

    for idx, f in enumerate(attachment):
      with open(f.path, "rb") as fp:
        part = MIMEApplication(fp.read())
      part.add_header("Content-Disposition", "attachment", filename=_encode(f.basename))
      part.add_header("content-ID", "<%s>" % idx)
      msg.attach(part)

    self.server.sendmail(self.fromAddr, [
      _decodeAddr(item)[1] for item in toAddr
    ], msg.as_string())
    return self



#!/usr/bin/env python
# coding: utf-8

from typing import Any, List, Tuple, Dict, Callable, TypeVar
import os, random, subprocess, codecs, json, socket, webbrowser, shutil, re, sys, math
from http.server import HTTPServer, SimpleHTTPRequestHandler
""" pillow
  如果没有pillow, 详见 https://pillow.readthedocs.io/en/stable/installation.html
  pip install pillow
"""
from PIL import Image

""" multiprocessing
  from multiprocessing import Pool

  pool = Pool(processes=4)

  def doSomething(i):
    print(i)

  for i in range(100):
    pool.apply_async(doSomething, args=(i,))

  pool.close()
  pool.join()
"""

""" arguments
  from argparse import ArgumentParser

  parser = ArgumentParser()
  parser.add_argument("-b", "--boolean", help="some help", action="store_true")
  parser.add_argument("-s", "--string", help="some help")
  parser.add_argument("-i", "--int", type=int, help="some help")

  args = parser.parse_args()

  print(args.string)
"""

""" 路径区别
  os.getcwd(): 程序执行目录(绝对路径)
  sys.path[0]: 主模块所在目录(绝对路径)
  sys.argv[0]: 主模块文件路径(绝对路径 | 相对路径)
  __file__: 作为入口时(__name__ == "__main__"), 是当前文件名, 作为模块被引入时, 是当前文件绝对路径
"""

def resolve(*args: str) -> str:
  """
  不同于 Node.js 中的 path.resolve, 此处并非返回绝对路径, 而是规范化的路径
  """
  return os.path.normpath(os.path.join(*args)).replace("\\", "/")

def createIfNotExists(dirPath: str) -> "Dir":
  if not os.path.isdir(dirPath):
    os.makedirs(dirPath)
  return Dir(dirPath)

def createHttpServer(host: str=socket.gethostbyname(socket.gethostname()), port: int=8080, webRoot: str=os.getcwd(), page: str="", open: bool=True):
  os.chdir(webRoot)
  httpd = HTTPServer((host, port), SimpleHTTPRequestHandler)
  url = "%s://%s:%s/%s" % ("http", host, port, re.sub(r'(\.\/|\/|\.\\|\\)', "", page))
  print("\nserver running on  %s\n" % url)
  if open:
    webbrowser.open(url)
  try:
    httpd.serve_forever()
  except KeyboardInterrupt:
    print("KeyboardInterrupt")
    pass
  
def question(q: str, defaultValue: str = None) -> str:
  qStr = "%s 默认值【%s】" % (q, defaultValue) if defaultValue != None else q
  if defaultValue == None:
    while True:
      inputStr = input(qStr).strip()
      if inputStr != "":
        return inputStr
  else:
    assert isinstance(defaultValue, str), "is not string: %s" % defaultValue
    inputStr = input(qStr).strip()
    if inputStr == "":
      return defaultValue
    else:
      return inputStr

def questionInt(q: str, defaultValue: int = None) -> int:
  qStr = "%s 默认值【%s】" % (q, defaultValue) if defaultValue != None else q
  if defaultValue == None:
    while True:
      try:
        return int(input(qStr).strip())
      except Exception:
        pass
  else:
    assert isinstance(defaultValue, int), "wrong type: %s is not int" % defaultValue
    while True:
      inputStr = input(qStr).strip()
      if inputStr == "":
        return defaultValue
      else:
        try:
          return int(inputStr)
        except Exception:
          pass

def questionFloat(q: str, defaultValue: float = None) -> float:
  qStr = "%s 默认值【%s】" % (q, defaultValue) if defaultValue != None else q
  if defaultValue == None:
    while True:
      try:
        return float(input(qStr).strip())
      except Exception:
        pass
  else:
    assert isinstance(defaultValue, float), "wrong type: %s is not float" % defaultValue
    while True:
      inputStr = input(qStr).strip()
      if inputStr == "":
        return defaultValue
      else:
        try:
          return float(inputStr)
        except Exception:
          pass

def questionBool(q: str, defaultValue: bool = False) -> bool:
  qStr = "%s【Y/n】 默认值【%s】" % (q, "Y" if defaultValue else "n")
  assert isinstance(defaultValue, bool), "wrong type: %s is not bool" % defaultValue
  while True:
    inputStr = input(qStr).strip().lower()
    if inputStr == "":
      return defaultValue
    if inputStr == "y":
      return True
    if inputStr == "n":
      return False

class Base(object):
  # 规范路径
  path: str
  # 父目录的路径
  dirname: str
  # 文件名(包括后缀)
  basename: str
  # 文件名(不包括后缀)
  name: str
  # 后缀
  suffix: str
  # 假的无后缀文件名(之所以说是假的, 是因为目录没有后缀)
  fakeName: str
  # 假的后缀(之所以说是假的, 是因为目录没有后缀)
  fakeSuffix: str
  def __init__(self, *p: str):
    super(Base, self).__init__()
    self.path: str = resolve(*p)

  @property
  def dirname(self) -> str:
    return os.path.dirname(self.path)

  @property
  def basename(self) -> str:
    return os.path.basename(self.path)

  @property
  def splitname(self) -> Tuple[str, str]:
    return os.path.splitext(self.basename)

  @property
  def name(self) -> str:
    return self.basename

  @property
  def suffix(self) -> str:
    return ""

  @property
  def fakeName(self) -> str:
    return self.splitname[0]

  @property
  def fakeSuffix(self) -> str:
    return self.splitname[1]

  @property
  def parent(self) -> "Base":
    return Base(self.dirname)

  def childOf(self, *p: str) -> "Base":
    return Base(self.path, *p)

  def siblingOf(self, *p: str) -> "Base":
    return Base(self.dirname, *p)

  @property
  def isImg(self) -> bool:
    _isImg = False
    try:
      with Image.open(self.path) as img:
        _isImg = True
    except Exception:
      pass
    return _isImg

  @property
  def isFile(self) -> bool:
    return os.path.isfile(self.path)
  
  @property
  def isDir(self) -> bool:
    return os.path.isdir(self.path)

  def toAbsPath(self) -> "Base":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

  def createAsDir(self) -> "Dir":
    if not self.isDir:
      os.makedirs(self.path)
    return Dir(self.path)
  
  def createAsFile(self) -> "File":
    if not self.isFile:
      self.parent.createAsDir()
      with open(self.path, "w") as f:
        pass
    return File(self.path)

  def asFile(self) -> "File":
    return File(self.path)

  def asDir(self) -> "Dir":
    return Dir(self.path)

class File(Base):
  def __init__(self, *p: str):
    super(File, self).__init__(*p)
    assert self.isFile, "is not file: %s" % self.path

  @property
  def name(self) -> str:
    return self.fakeName

  @property
  def suffix(self) -> str:
    return self.fakeSuffix

  @property
  def parent(self) -> "Dir":
    return Dir(self.dirname)
  
  def read(self, encoding="UTF-8") -> str:
    result = ""
    with codecs.open(self.path, "r", encoding=encoding) as f:
      result = f.read()
    return result

  def __write(self, s: str, mode: str, encoding="UTF-8") -> "File":
    with codecs.open(self.path, mode, encoding=encoding) as f:
      f.write(s)
    return self

  def write(self, s: str, encoding="UTF-8") -> "File":
    return self.__write(s, "w", encoding=encoding)
  
  def aWrite(self, s: str, encoding="UTF-8") -> "File":
    return self.__write(s, "a", encoding=encoding)
  
  def toAbsPath(self) -> "File":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

class Dir(Base):
  def __init__(self, *p: str):
    super(Dir, self).__init__(*p)
    assert self.isDir, "is not dir: %s" % self.path

  @property
  def parent(self) -> "Dir":
    return Dir(self.dirname)

  @property
  def fileNames(self) -> List[str]:
    return [f for f in os.listdir(self.path) if os.path.isfile(resolve(self.path, f))]
  
  @property
  def files(self) -> List["File"]:
    return [File(self.path, f) for f in self.fileNames]
  
  @property
  def dirNames(self) -> List[str]:
    return [d for d in os.listdir(self.path) if os.path.isdir(resolve(self.path, d))]
  
  @property
  def dirs(self) -> List["Dir"]:
    return [Dir(self.path, d) for d in self.dirNames]
  
  @property
  def allDirs(self) -> List["Dir"]:
    dirs = [d for d in self.dirs]
    curIdx = 0
    while curIdx < len(dirs):
      dirs.extend(dirs[curIdx].dirs)
      curIdx += 1
    return dirs

  @property
  def allFiles(self) -> List["File"]:
    files = [f for f in self.files]
    for d in self.allDirs:
      files.extend(d.files)
    return files

  def includes(self, dirA: "Dir") -> bool:
    # 当不含有dirA中的文件时，有print提示；
    # 当不含有dirA中的目录时，没有提示
    for f in dirA.files:
      if f.basename not in self.fileNames:
        print("【%s】 not in 【%s】" % (f.basename, self.path))
        return False
    for d in dirA.dirs:
      if d.basename not in self.dirNames:
        return False
      if not Dir(self.path, d.basename).includes(d):
        return False
    return True
  
  def toAbsPath(self) -> "Dir":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

class Img(File):
  """
  很不成熟, 尽量别用
  除非你真的知道你在干嘛
  """
  obj: Image.Image
  def __init__(self, *p: str):
    super(Img, self).__init__(*p)
    self.obj = Image.open(self.path)
    assert self.isImg, "is not img: %s" % self.path
  
  @property
  def isJpg(self) -> bool:
    # return self.obj.mode == "RGB" or (self.suffix.lower() == ".jpg" and self.obj.mode.lower() == "p")
    return self.suffix.lower() in [".jpg", ".jpeg"]

  @property
  def isPng(self) -> bool:
    # return self.obj.mode == "RGBA" or (self.suffix.lower() == ".png" and self.obj.mode.lower() == "p")
    return self.suffix.lower() == ".png"

  def close(self):
    self.obj.close()

  def resize(self, size: Tuple[int, int], resample=Image.LANCZOS) -> "Img":
    self.obj = self.obj.resize(size, resample=resample)
    return self
  
  def toAbsPath(self) -> "Img":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

  # args 是 Image.save 的参数
  def saveAs(self, p="", **args):
    self.obj.save(p or self.path, **args)

class Jpg(Img):
  def __init__(self, *p: str):
    super(Jpg, self).__init__(*p)
    assert self.isJpg, "is not jpg: %s, mode: %s" % (self.path, self.obj.mode)
  
  def saveAs(self, p="", optimize=True, quality=90, progressive=True, subsampling=1) -> "Jpg":
    self.obj.save(p or self.path, format="JPEG", optimize=optimize, quality=quality, progressive=progressive, subsampling=subsampling)
    return self
  
  # quality = 0 时，不压缩
  def saveToPng(self, p="", quality=0) -> str:
    p = p or self.path
    base = Base(p)
    base.parent.createAsDir()
    suffix = ".png"
    name = base.fakeName if base.fakeSuffix.lower() == suffix else base.basename
    targetPath = resolve(base.dirname, "%s%s" % (name, suffix))
    rgba = self.obj.convert("RGBA")
    rgba.save(targetPath)
    if quality > 0:
      Png(targetPath).saveAs(quality=quality)
    return targetPath

  def toAbsPath(self) -> "Jpg":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

class Png(Img):
  def __init__(self, *p: str):
    super(Png, self).__init__(*p)
    assert self.isPng, "is not png: %s, mode: %s" % (self.path, self.obj.mode)

  def saveAs(self, p="", quality=90) -> "Png":
    tempRoot = Base("/.tangUtils-pngquant-temp").createAsDir()
    p = p or self.path
    qualityStr = "%s-%s" % (quality, quality + 2)
    newPath = tempRoot.childOf("%s-%s.png" % (str(random.random()), str(random.random()))).path
    self.obj.save(newPath)
    subprocess.call("pngquant.exe --force --ext=.png --quality=%s %s" % (qualityStr, newPath))
    shutil.move(newPath, p)
    return self

  def saveToJpg(self, p="", optimize=True, quality=94, progressive=True, subsampling=1) -> str:
    p = p or self.path
    base = Base(p)
    base.parent.createAsDir()
    suffix = ".jpg"
    name = base.fakeName if base.fakeSuffix.lower() == suffix else base.basename
    targetPath = resolve(base.dirname, "%s%s" % (name, suffix))
    rgb = self.obj.convert("RGB")
    rgb.save(targetPath, optimize=optimize, quality=quality, progressive=progressive, subsampling=subsampling)
    return targetPath

  def toAbsPath(self) -> "Png":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

class Json(File):
  def readAsJson(self) -> Any:
    s = super(Json, self).read()
    return json.loads(s)

  # indent为零时自动转为None(即不支持indent == 0, 我认为indent == 0不常用, 还不如让indent == None用得舒服点)
  def writeAsJson(self, obj, indent=2, ensure_ascii=False) -> "Json":
    with open(self.path, "w", encoding="UTF-8") as f:
      json.dump(obj, f, indent=indent if indent else None, ensure_ascii=ensure_ascii)
    return self

  def toAbsPath(self) -> "Json":
    self.path = os.path.abspath(self.path).replace("\\", "/")
    return self

def getInputFiles() -> List["File"]:
  files = []
  for arg in sys.argv[1:]:
    base = Base(arg)
    if base.isFile:
      files.append(File(arg))
    elif base.isDir:
      for f in Dir(arg).allFiles:
        files.append(f)
  return files

Callback = Callable[[], None]

class Cmd:
  prompt: str
  callback: Callback
  next: List["Cmd"]

  def __init__(self, prompt: str, callback: Callback=None, next: List["Cmd"]=[]):
    super().__init__()
    self.prompt = prompt
    self.callback = callback
    self.next = next

def runCmdList(cmdList: List[Cmd]) -> bool:
  """
  返回值表示是否正常退出;
  True为正常退出;
  False表示非正常退出, 如"命令列表为空"或者"ctrl+c"退出等
  """
  if len(cmdList) == 0:
    print("命令列表为空")
    return False
  cmdMap: Dict[int, Cmd] = {}
  prompts = ["", "--- 可执行命令列表 ---"]
  i = 0
  for cmd in cmdList:
    i += 1
    cmdMap[i] = cmd
    prompts.append("%s: %s" % (i, cmd.prompt))
  prompts.append("------")
  promptStr = "\n".join(prompts) + "\n"
  result = -1
  try:
    result = questionInt(promptStr)
  except KeyboardInterrupt:
    print("您已取消输入")
    return False
  while result not in cmdMap:
    try:
      result = questionInt("请输入有效命令")
    except KeyboardInterrupt:
      print("您已取消输入")
      return False
  if cmdMap[result].callback:
    cmdMap[result].callback()
  if len(cmdMap[result].next) > 0:
    runCmdList(cmdMap[result].next)
  return True

Size = Tuple[int, int]

def resizeWithin(origin: Size, target: Size) -> Size:
  """
  originW / originH == targetW / targetH
  """
  originW, originH = origin
  targetW, targetH = target
  if 0 in [originW, originH, targetW, targetH]:
    raise Exception("尺寸不得含有0")
  if originW < 0 or originH < 0:
    raise Exception("源尺寸必须均大于0")
  # 不允许是其他负数, 是为了防止使用者误输入
  if (targetW < 0 and targetW != -1) or (targetH < 0 and targetH != -1):
    raise Exception("目标尺寸不得为其他负数 (但可以为-1, 表示以另一个正数为标准, 保持比例缩放)")
  if targetW < 0 and targetH < 0:
    raise Exception("目标尺寸不得同时为负数")

  aspect = originW / originH

  if targetW < 0:
    return (math.floor(aspect * targetH), targetH)
  if targetH < 0:
    return (targetW, math.floor(targetW / aspect))
  if aspect < targetW / targetH: # 瘦高, 以高为准
    return (math.floor(aspect * targetH), targetH)
  # 否则以宽为准
  return (targetW, math.floor(targetW / aspect))



#!/usr/bin/env python
# coding: utf-8

""" doc
  see
    https://zhuanlan.zhihu.com/p/26159930
    https://pypi.org/project/twine/

  # 执行下列命令前, 需要先更新版本号
  1. python setup.py sdist
  2. twine upload dist/tangUtils-0.1.0.tar.gz
"""

from setuptools import setup

setup(
  name='tangUtils',
  version='0.1.0',
  author='xiaomingtang',
  author_email='1038761793@qq.com',
  url='https://github.com/xiaomingTang/python-utils',
  description=u'文件/目录相关utils',
  packages=['tangUtils'],
  install_requires=[
    "pillow"
  ],
  entry_points={
    'console_scripts': []
  }
)


# encoding='utf-8'
from typing import List
import os, shutil, re, sys
from ftplib import FTP
from tangUtils.main import resolve, Base, File, Dir, question

def toLatin(p: str):
  return p.encode("utf-8").decode("latin1")

class TangFtp(object):
  host: str
  port: int
  username: str
  password: str
  remoteRoot: str
  localRoot: str
  ftp: FTP = None

  def __init__(self, host="", port=21, username="", password="", remoteRoot="/", localRoot="/"):
    super().__init__()
    self.host = host
    self.port = port
    self.username = username
    self.password = password
    self.remoteRoot = remoteRoot
    self.localRoot = localRoot
    self.__login()

  def __login(self) -> "TangFtp":
    self.__logout()
    self.ftp = FTP()
    self.ftp.connect(host=self.host, port=self.port)
    self.ftp.login(user=self.username, passwd=self.password)
    return self

  def __logout(self) -> "TangFtp":
    if self.ftp:
      self.ftp.close()
    self.ftp = None
    return self

  def __createDirIfNotExists(self, remoteDirPath: str):
    oldDir = self.ftp.pwd()
    if remoteDirPath[0] == "/":
      self.ftp.cwd("/")
    for subName in remoteDirPath.split("/"):
      if bool(subName):
        try:
          self.ftp.cwd(subName)
        except Exception:
          self.ftp.mkd(subName)
          self.ftp.cwd(subName)
    self.ftp.cwd(oldDir)

  def __ftpDownload(self, fromRemotePath: str, toLocalPath: str) -> bool:
    buffSize = 1024
    success = True
    Base(toLocalPath).createAsFile()
    with open(toLocalPath, "wb") as fp:
      try:
        self.ftp.retrbinary("RETR %s" % toLatin(fromRemotePath), fp.write, buffSize)
        self.ftp.set_debuglevel(0)
      except Exception as e:
        print("【ftp download error】")
        print(e)
        success = False
    return success

  def __ftpUpload(self, fromLocalPath: str, toRemotePath: str) -> bool:
    self.__createDirIfNotExists(toLatin(Base(toRemotePath).parent.path))
    buffSize = 1024
    success = True
    # 这儿本地文件必须存在, 本来就是上传文件, 本地没有文件你上传什么? 所以就没判断
    with open(fromLocalPath, "rb") as fp:
      try:
        self.ftp.storbinary("STOR %s" % toLatin(toRemotePath), fp, buffSize)
        self.ftp.set_debuglevel(0)
      except Exception as e:
        print("【ftp upload error】")
        print(e)
        success = False
    return success

  def upload(self, s: List[Base]):
    total = len(s)
    for i, base in enumerate(s):
      print("%s / %s" % (i + 1, total), end="\r")
      relPath = os.path.relpath(base.path, self.localRoot)
      # 两点或者斜杠开头, 都不在目录中
      if re.match(r"^(\.\.|\\|\/)", relPath):
        print("--%s--%s--" % (base.path, relPath))
        print("文件错误: 【%s】不在根目录【%s】中" % (base.path, self.localRoot))
        return self
      remotePath = resolve(self.remoteRoot, relPath)
      if base.isFile:
        self.__ftpUpload(base.path, remotePath)
      elif base.isDir:
        self.upload(Dir(base.path).allFiles)
    return self

  def download(self, s: List[Base]):
    total = len(s)
    for i, base in enumerate(s):
      print("%s / %s" % (i + 1, total), end="\r")
      relPath = os.path.relpath(base.path, self.localRoot)
      # 两点或者斜杠开头, 都不在目录中
      if re.match(r"^(\.\.|\\|\/)", relPath):
        print("文件错误: 【%s】不在根目录【%s】中" % (base.path, self.localRoot))
        return self
      remotePath = resolve(self.remoteRoot, relPath)
      if base.isFile:
        self.__ftpDownload(remotePath, base.path)
      elif base.isDir:
        self.download(Dir(base.path).allFiles)
    return self

if __name__ == "__main__":
  try:
    ftp = TangFtp()
    tar = [Base(p) for p in sys.argv[1:]]
    ftp.upload(tar)
  except Exception as e:
    print(e)
    question("输入 任意字符 以退出", "Y")
